from argparse import ArgumentParser, Action
from os import PathLike
from pathlib import Path

from yaml import safe_load


class ASNViewerConfig:
    """
    Configuration container class for the ASN Viewer
    """
    def __init__(self, config_file: PathLike):
        self.__parse_config_file(config_file)


    def __parse_config_file(self, config_file: PathLike):
        """
        Parses configuration file
        :param config_file: Path to config file

        """
        with open(config_file, 'r') as config_file:
            self.raw_yaml_config = safe_load(config_file)
            self.definition = Path(self.raw_yaml_config['definition'])
            self.json_output_directory = Path(self.raw_yaml_config['output'])
            self.object_name = self.raw_yaml_config['object-name']
            self.search = self.raw_yaml_config['search']
            self.files = tuple(Path(f) for f in self.raw_yaml_config['files'])
            self.output = self.raw_yaml_config['output']

    def __str__(self):
        return (f'asn-schema-file: {self.definition}\n'
                f'output: {self.json_output_directory}\n'
                f'object-name: {self.object_name}\n'
                f'search {self.search}\n')

class ParseSearchKwargs(Action):
    """
    ASN Viewer Argument parser helper Action for search params

    """
    def __call__(self, parser, namespace, values, option_string=None, ):
        setattr(namespace, self.dest, dict())
        for value in values:
            key, value = value.split('=')
            getattr(namespace, self.dest)[key] = tuple(s.strip() for s in value.replace(';', ',').split(','))


class ASNViewerCmdArguments(ArgumentParser):
    """
    ASN Viewer Argument parser

    """
    def __init__(self):
        super().__init__()

        self.add_argument('-c', '--config', type=Path, help='Config filename')
        self.add_argument('-d', '--definition', type=Path, help='Schema Definition file')
        self.add_argument('-o', '--output', type=Path, help='Output filename')
        self.add_argument('-n', '--object-name', help='ASN Object Name')
        self.add_argument('-f', '--files', nargs='*', help='Filenames to decode')

        self.add_argument('-s', '--search',
                          nargs='*', action=ParseSearchKwargs,
                          help='Search filter: fieldName=value. Multiple values per key should be separated by a comma')


from pathlib import Path

from .asn_decoder import ASNDecoder
from .asn_viewer_config import ASNViewerConfig, ASNViewerCmdArguments


def main():
    try:
        args = ASNViewerCmdArguments().parse_args()

        conf = ASNViewerConfig(Path(args.config)) if args.config else None

        decoder = ASNDecoder(args.definition or conf.definition, args.object_name or conf.object_name)
        decoder.load_files(args.files or conf.files or tuple())

        if args.output or conf:
            decoder.save_decoded_to_file(Path(args.output or conf.output), args.search or conf.search or None)
        else:
            decoder.print_file_data_json(args.search or conf.search or None)

    except AttributeError as e:
        print(f'Missing parameter: {e.name.replace("_", "-")}')
        exit(1)
    except KeyError as ke:
        print(f'Field name {ke} is not in the ASN object definition')
        exit(2)


from setuptools import setup, find_packages

with open('./package_description.rst', 'rt', encoding='utf8') as readme, \
     open('./requirements.txt', 'rt', encoding='utf8') as reqs, \
     open('./LICENSE', 'rt', encoding='utf-8') as lic:
    setup(
        name='asn_viewer',
        maintainer='Sergey Yakimov',
        maintainer_email='sergwy@gmail.com',
        version='0.0.2',
        url='https://gitlab.com/sergwy/asn-viewer',
        description='ASN decoder and viewer',
        long_description_content_type='text/x-rst',
        long_description=readme.read(),
        packages=find_packages(),
        license=lic.read(),
        install_requires=[r for r in reqs],
        entry_points={
            'console_scripts': [
                'asn-viewer = asn_viewer.runapp:main'
            ]
        }
    )



from json import dumps
from numbers import Number
from os import PathLike
from typing import List, Set, Any, Generator, AnyStr

from asn1tools import parse_files
from asn1tools.compiler import Specification
from asn1tools.codecs import ber, type_checker, constraints_checker


class ASNDecoder(Specification):
    """
    ASN Decoder class
    Extends asn1tools.compiler.Specification
    """
    def __init__(self, spec_file: PathLike[str], object_name: str):
        self.object_name = object_name
        self.schema = parse_files([spec_file])
        self.decoded_asn_objects = []
        super().__init__(ber.compile_dict(self.schema, True),
                         ber.decode_full_length,
                         type_checker.compile_dict(self.schema, True),
                         constraints_checker.compile_dict(self.schema, True)
                         )

    def jsonify(self, data: bytes):
        return dumps(self.decode(self.object_name, data))

    @staticmethod
    def __values_match(a: str, b: Any) -> bool:
        try:
            if isinstance(b, int):
                return int(a) == b
            elif isinstance(b, float):
                return float(a) == b
            elif isinstance(b, str):
                return a == b
        except ValueError:
            return False


    def __decode_files(self, files: List[PathLike[str]]):
        for f in  files:
            try:
                with open(f, 'rb') as asn_object_b:
                    while b0 := asn_object_b.read(1):
                        b1 = asn_object_b.read(1)
                        b2 = asn_object_b.read(int.from_bytes(b1) - 0x80)
                        b3 = asn_object_b.read(int.from_bytes(b2))
                        yield self.decode(self.object_name, b0 + b1 + b2 + b3)
            except FileNotFoundError:
                # Don't care if the source file is not accessible
                continue

    def load_files(self, files: List[PathLike]):
        for r in self.__decode_files(files):
            self.decoded_asn_objects.append(r)

    def save_decoded_to_file(self, filename: PathLike[str], search_filter: dict | None = None):
        with open(filename, 'wt') as target_file:
            asn_objects = self.decoded_asn_objects if not search_filter else self.__search_data(**search_filter)
            target_file.write(
               "\n".join([dumps(u) for u in asn_objects])
            )

    def __search_data(self, **kwargs) -> list[dict]:
        return list(
            u for u in self.decoded_asn_objects
            if all(any(self.__values_match(w, u[k]) for w in v) for k, v in kwargs.items())
        )

    def print_search_result(self, **kwargs):
        print(self.__search_data(**kwargs))

    def print_file_data_json(self, search_filter: dict | None):
        asn_objects = self.decoded_asn_objects if not search_filter else self.__search_data(**search_filter)
        print(*(dumps(o) for o in asn_objects))
